//Copyright statements of copy-pasted code segments
/*
  Rui Santos
  Complete project details at https://RandomNerdTutorials.com/esp-now-two-way-communication-esp32/
  
  Simplified for demo purposes by R Hart 5/18/20.
  
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files.
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
*/

//Required libraries
#include <esp_now.h>
#include <WiFi.h>
#include <Adafruit_NeoPixel.h>
#include <Preferences.h>

//Pin definitions
#define EngLDir     32 //Dir of left motor
#define EngLPWM     33 //PWM of left motor
#define EngRDir     25 //Dir of right motor
#define EngRPWM     26 //PWM of right motor
#define SPDL        34 //Speed sense left
#define SPDR        22 //Speed sense right
#define Fan         18 //Fan control
#define LED         19 //LED control
#define RGB         23 //RGB data
#define VSense      35 //Voltage sense
#define TSense0     36 //Temperature sense #1
#define TSense1     39 //Temperature sense #2
#define WM          27 //Weapon motor
#define WSM1        15 //Weapon support motor 1 (fire)
#define WSM2        21 //Weapon support motor 2 (reload)
#define WS1         13 //Weapon sensor 1 (belt status)
#define WS2         14 //Weapon sensor 2 (drum status)

//Variable definitions
const int freq          = 20000;  //PWM frequency: 20 kHz
const int EngPWM_R      = 0;      //PWM channel 0: right track motor
const int EngPWM_L      = 1;      //PWM channel 1: left track motor
const int LEDbrightness = 2;      //PWM channel 2: headlight LEDs
const int Fanchannel    = 3;      //PWM channel 3: cooling fans
const int Weapon        = 4;      //PWM channel 4: weapon motors (both on one channel)
const int BeltMotor     = 5;      //PWM channel 5: weapon belt motor
const int DrumMotor     = 6;      //PWM channel 6: weapon drum motor
const int resolution    = 8;      //PWM is used with 8 bit resolution (duty cycle: 0 to 255)

#define RGBcount 22 //number of RGB LEDs of the strip

int LEDSpeed        = 6; //change this value to adjust the speed of the LED running light
int counter         = 5;
int IntCount        = 0;
int currentIntG     = 0;
int currentIntR     = 0;
int currentIntB     = 0;
int IntMaxG         = 0;
int IntMaxR         = 0;
int IntMaxB         = 0;
int wave            = 0;
int time0           = 0;
int time1           = 1;
int drive           = 0;
int runvar          = 0;
int vbat            = 0;
int fantimer        = 500;
int signalquality   = 1;
int signalout       = 0;
int onoff           = 0;
int rgbtimer        = 0;
int batlim          = 0;
int spdlv           = 0;
int spdlv_old       = 0;
int spdrv           = 0;
int spdrv_old       = 0;
int spdtime0        = 0;
int spdtime1        = 0;
int loopcount       = 0;
int lspd            = 0;
int rspd            = 0;
int pwml            = 0;
int pwmr            = 0;
int dirl            = 0;
int dirr            = 0;
int defaultheight   = 0;
int autoadjust      = 0;
int spderrR         = 0;
int spderrL         = 0;
bool error          = 0;
int wpnspd          = 0;
int trigger         = 0;
bool belt           = 0;
int drum            = 0;
int wpncounter      = 0;
int pos             = 0;
bool drummove       = 0;
int weapondebounce  = 0;
int weapontimer0    = 0;
int weapontimer1    = 0;

Adafruit_NeoPixel strip(RGBcount, RGB, NEO_GRB + NEO_KHZ800);

//ESPNOW setup
//MAC addresses: Enter the specific MAC addresses of your ESPs to establish the communication
uint8_t broadcastAddress[] =  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};    //MAC address of remote control

// Define a data structure for sending data to the remote control
typedef struct send_message {
  int voltage;
  int errorcode;
  int drumpos;
} send_message;
 
// Create a structured object
send_message outgoing;

// Define a data structure for receiving information from the remote control
typedef struct receive_message {
  int spdL;
  int spdR;
  int hv;
  int R;
  int G;
  int B;
  int wpn;
  int ls;
  int drv;
} receive_message;
 
// Create a structured object
receive_message incoming;

// Callback when data is sent
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  if (status == ESP_NOW_SEND_SUCCESS&&signalquality<19){ //signal quality is measured by changing the value of "signalquality" according to how many data packages were not received
    signalquality = signalquality +1;
  }else if(signalquality>0){
    signalquality = signalquality -1;
  }  
}

// Callback when data is received
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  memcpy(&incoming, incomingData, sizeof(incoming));
}

esp_now_peer_info_t peerInfo; //make sure that the ESP-Now lines are in the exact same order as in this sketch. Otherwise, there may be malfunctions.

//preferences library allows you to store variable values within the flash so that their value can be retrieved on startup. Used to store user-adjustable settings.
Preferences preferences;

void setup() {
  preferences.begin("Defaults", false);
  defaultheight = preferences.getInt("Height", 30); //the last height value is saved to avoid the tank to adjust its height to a default value every single time it is started
  autoadjust = preferences.getInt("Adjust",0); //commands for auto or manual tilt adjustment are only sent to the tank when changing the settings. Thus, the last setting needs to be available on startup
  pos = preferences.getInt("Ammunition", 0); //the tank stores the ammunition count so that it keeps its bullet count when switching it off and on
  
  //initialize serial monitor - can be used for code debugging
  Serial.begin(115200);
  //initialize serial 2 - used to forward commands to the secondary µC
  Serial2.begin(9600, SERIAL_8N1, 16, 17);
  
  //Set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);

  //Initialize ESP-NOW
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  
  // Register peer
  memcpy(peerInfo.peer_addr, broadcastAddress, 6);
  peerInfo.channel = 0;  
  peerInfo.encrypt = false;
  
  // Add peer        
  if (esp_now_add_peer(&peerInfo) != ESP_OK){
    Serial.println("Failed to add peer");
    return;
  }
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);

  pinMode(EngLDir, OUTPUT);
  pinMode(EngLPWM, OUTPUT);
  pinMode(EngRDir, OUTPUT);
  pinMode(EngRPWM, OUTPUT);
  pinMode(LED, OUTPUT);
  pinMode(SPDL, INPUT);
  pinMode(SPDR, INPUT);
  pinMode(WM, OUTPUT);
  pinMode(WSM1, OUTPUT);
  pinMode(WSM2, OUTPUT);
  pinMode(WS1, INPUT); //this is a photoelectric switch - no pullup resistor needed
  pinMode(WS2, INPUT_PULLUP); //this is classical limit switch - it needs a pullup resistor
  digitalWrite(EngLDir, LOW);
  digitalWrite(EngLPWM, LOW);
  digitalWrite(EngRDir, LOW);
  digitalWrite(EngRPWM, LOW);
  digitalWrite(LED, LOW);
  digitalWrite(WM, LOW);
  digitalWrite(WSM1, LOW);
  digitalWrite(WSM2, LOW);
  
  ledcSetup(EngPWM_R, freq, resolution); //setup of PWM control: PWM-Channel, frequency, resolution (here: both track motors)
  ledcAttachPin(EngRPWM, EngPWM_R); //Pin attachment: Pin name, PWM-Channel
  ledcSetup(EngPWM_L, freq, resolution);
  ledcAttachPin(EngLPWM, EngPWM_L);
  ledcWrite(EngPWM_R, 0); //write new status to PWM-Channel: PWM-Channel, value (0-255)
  ledcWrite(EngPWM_L, 0);
  
  ledcSetup(LEDbrightness, freq, resolution); //headlight LEDs
  ledcAttachPin(LED, LEDbrightness);
  ledcWrite(LEDbrightness, 0);

  ledcSetup(Fanchannel, freq, resolution); //cooling fans
  ledcAttachPin(Fan, Fanchannel);
  ledcWrite(Fanchannel, 255); //the fans spin up during startup as a part of the initialization procedure

  ledcSetup(Weapon, freq, resolution); //main weapon motors
  ledcAttachPin(WM, Weapon);
  ledcWrite(Weapon, 0);

  ledcSetup(BeltMotor, freq, resolution); //weapon belt motor
  ledcAttachPin(WSM1, BeltMotor);
  ledcWrite(BeltMotor, 0);

  ledcSetup(DrumMotor, freq, resolution); //weapon drum motor
  ledcAttachPin(WSM2, DrumMotor);
  ledcWrite(DrumMotor, 0);

  //RGB LED strip start
  strip.begin();
  strip.show(); //initialize all pixels to 'off'
  for (int i=0; i<RGBcount/2; i++){ //testing the strip: Run through all pixels with R, then G, then B
    strip.clear();
    strip.setPixelColor(i,127,0,0);
    strip.setPixelColor(i+11,127,0,0);
    strip.show();
    delay(50);
  }
  for (int i=0; i<RGBcount/2; i++){
    strip.clear();
    strip.setPixelColor(i,0,127,0);
    strip.setPixelColor(i+11,0,127,0);
    strip.show();
    delay(50);
  }
  for (int i=0; i<RGBcount/2; i++){
    strip.clear();
    strip.setPixelColor(i,0,0,127);
    strip.setPixelColor(i+11,0,0,127);
    strip.show();
    delay(50);
  }
  strip.clear(); //RGB LED strip is off until it receives input from the remote control
  strip.show();
    
  delay(2000);
  ledcWrite(Fanchannel, 0); //cooling fans stop after around 3.5 s (delay plus time for RGB LEDs
  digitalWrite(EngLDir, HIGH); //default directions for the tracks
  digitalWrite(EngRDir, LOW); 

  time0 = millis();
  outgoing.errorcode = 0; //default error: 0 (means no error detected)
}
 
void loop() {
  //the speeds of the two track motors (detected at the sun gears of the epicyclic gearboxes using sensor wheel & photoelectric switch
  //interrupts were tried but were not reliable (too low/unreliable counts). The loop is fast enough to detect input counts reliably using this option
  if (spdlv_old != digitalRead(SPDL)){
    spdlv_old = digitalRead(SPDL);
    spdlv     = spdlv + 1;
  }
  if (spdrv_old != digitalRead(SPDR)){
    spdrv_old = digitalRead(SPDR);
    spdrv     = spdrv + 1;
  }
  
  spdtime0 = millis(); //delay & denoise for stall detection: speed level is the number of counts per 100 loop iterations
  if (spdtime0>spdtime1+99){
    lspd     = spdlv;
    rspd     = spdrv;
    spdlv    = 0;
    spdrv    = 0;
    spdtime1 = millis();
  }

  if (error==1){ //safety stop of all motors and actuators when an error is encountered
    pwml = 0;
    pwmr = 0;
    ledcWrite(EngPWM_L, 0);
    ledcWrite(EngPWM_R, 0); 
    ledcWrite(Fanchannel,0);
    ledcWrite(LEDbrightness, 0);
    ledcWrite(Weapon, 0);
    ledcWrite(BeltMotor, 0);
    ledcWrite(DrumMotor, 0);
    Serial2.write(255); //255 tells the secondary µC to stop the linear actuators
    esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing));
    RGBwarning(); //this function lets the RGB LED strip indicate that there is an error (and which error)
  }else{ //only enter the signal quality check&feedback part when there is no error
    if (signalquality==0&&signalout==0){ //emergency stop of the track motors when the signal is lost
      pwml = 0;
      pwmr = 0;      
      ledcWrite(EngPWM_L, 0);
      ledcWrite(EngPWM_R, 0);
      ledcWrite(Fanchannel,0);
      signalout = 1;
    }else if (signalquality>10&&signalout==1){ //signal is back after it was lost
      signalout = 0;
      strip.clear();
      onoff = 0;
      for (int i=0; i<RGBcount/2; i++){ //let every second RGB LED blink green once for 1 s when the signal is back
        if (onoff==0){
          strip.setPixelColor(i,0,127,0);
          strip.setPixelColor(i+11,0,127,0);
          onoff = 1;  
        }else{       
          onoff = 0;
        }     
      }
      strip.show();
      delay(1000);    
    }else if(signalout==1){ //signal is (still) lost: indicate by red blinking of every second RGB LED  
      strip.clear();
      onoff=0;
      if (rgbtimer<17){ //the RGB timer is responsible for the blinking frequency of the LEDs during signal loss condition (around 1 s for one on/off cycle)
        for (int i=0; i<RGBcount/2; i++){ //red blinking of every second LED
          if (onoff==0){
            strip.setPixelColor(i,127,0,0);
            strip.setPixelColor(i+11,127,0,0);
            onoff=1;  
          }else{
            onoff=0;
          }     
        }            
      }
      strip.show();      
    }else if(signalout==0){ //weapon control and speed control are only entered when there is no hardware error and no signal loss
      WpnCntrl(); //could probably also be shifted the following if-condition to increase the loop time (but I didn't try that)
      loopcount = loopcount+1;
      if (loopcount>99){ //the main loop has to be fast for speed detection -> the speed update is only done every 100th iteration
        if (incoming.drv == 0){ //indicates the drive mode: either direct or interpolated; calls the right function accordingly
          updatespeed0();  
        }else{
          updatespeed1();
        }        
        loopcount = 0;
      }    
    }
  }
  
  vbat = vbat + analogRead(VSense); //battery voltage sensing routine: entered always (irrespectively of the signal connection) unless an error occurred
  if (runvar>99){ //the value is averaged 100 times to reduce noise
    runvar = 0;
    outgoing.voltage = vbat/100; //the ADC value is sent without conversion to voltage - that is done in the remote control
    if (outgoing.voltage<2247){ //2247 equals around 11.0V (a bit less than 3.7V per cell of the 3S LiPo) - has to be measured manually for every ESP32!
      if (batlim<100){ //threshold counter for low battery error - is evaluated in the sentinel function
        batlim = batlim+1;
      }
    }else if (batlim>0&&batlim<99){
      batlim = batlim-1;
    }
    vbat = 0;
  }
  runvar = runvar+1;

  sentinel(); //the sentinel function evaluates inputs to identify if errors are existing
  
  time1 = millis();
  if (time1>(time0+30)&&error==0){ //time dependency: counting up the RGB timer and updating the RGB LEDs is done every 30 ms   
    fantimer = fantimer+1; //the fan timer is used to count time for the cooling fans (they turn on when driving and turn off 15 s after the last track movement)
    if (rgbtimer<34){
      rgbtimer = rgbtimer+1;
    }else{
      rgbtimer = 0;
    }
    if (signalout==0){
      updateRGB();
      ledcWrite(LEDbrightness, incoming.ls);               
    }
    esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing)); //sending data package every 30 ms
    if (signalout==1){ //stop height adjustments, the weapon, and switch of the LED headlights when the signal is lost
      Serial2.write(255);
      ledcWrite(LEDbrightness, 0);
      ledcWrite(Weapon, 0);
      wpnspd = 0;
    }else if (incoming.hv>0){ //the height value contains the desired height, auto- or manual tilt adjustment settings, or tilt adjustment commands when in manual mode
      if (incoming.hv==110){ //110 means automatic tilt adjustment
        autoadjust = 1;
        preferences.putInt("Adjust", autoadjust); //saved as a preference since it is only transmitted when actively changing this setting
      }else if (incoming.hv==111){ //111 means manual tilt adjustment
        autoadjust = 0;
        preferences.putInt("Adjust", autoadjust);
      }
      if (autoadjust==0||(lspd==0&&rspd==0)){ //tilt and height adjustments in automatic mode are only executed when the tank is standing still (always executed in manual mode)
        Serial2.write(incoming.hv);
      }else{
        Serial2.write(255); //command for secondary µC to stop linear actuators
      }
      if (incoming.hv<100){ //values above 0 and below 100 correspond to height values (but not tilt commands or setting changes)
      defaultheight = incoming.hv;
      preferences.putInt("Height", defaultheight); //store the last height value as a preference for next startup
      }
    }else{ //this else term can likely be deleted without effect on the functions
      Serial2.write(defaultheight); //send the last height input to the secondary µC if there is no further input but the signal loss not yet triggered
    }     
    time0 = millis();  
  }
  
}

//weapon control function - called every loop iteration to not miss a "fire" command event
void WpnCntrl(){
  outgoing.drumpos = pos;
  wpncounter = wpncounter+1;
  if (wpncounter==20){ //weapon speed adjustment is done smoothly by incrementally changing the PWM duty cycle (update every 20th iteration)
    wpncounter = 0;
    if (incoming.wpn==4){ //conversion of transmitted weapon speed levels (0=off; 1-4=on [but different duty cycles])
      if(wpnspd<255){ 
        wpnspd = wpnspd+1; //incremental change of PWM duty cycle
      }
    }else if (incoming.wpn==3){
      if(wpnspd<240){
        wpnspd = wpnspd+1;
      }else if(wpnspd>240){
        wpnspd = wpnspd-1;
      }
    }else if (incoming.wpn==2){
      if(wpnspd<225){
        wpnspd = wpnspd+1;
      }else if(wpnspd>225){
        wpnspd = wpnspd-1;
      }
    }else if (incoming.wpn==1){
      if(wpnspd<210){ //this relatively high min weapon speed is needed to avoid bullets from getting stuck between the acceleration wheels (and stalling them)
        wpnspd = wpnspd+1;
      }else if(wpnspd>210){
        wpnspd = wpnspd-1;
      }        
    }else if (incoming.wpn==0){
      if(wpnspd>0){
        wpnspd = wpnspd-1;
      }
    }
  }
  if (pos<17){
    ledcWrite(Weapon,wpnspd); //update the weapon motor PWM duty cycle if the weapon is not jammed
  }else{
    ledcWrite(Weapon,0); //safety stop when the weapon is jammed
  }

  if (incoming.wpn==6){ //reload command: reset the drum position
    pos = 0;
    preferences.putInt("Ammunition", pos); //save ammunition count as a preference after reloading
  }
  
  belt = digitalRead(WS1); //read the sensors of belt and drum
  drum = digitalRead(WS2);

  if (incoming.wpn==5&&wpnspd>150&&pos<16&&trigger==0){ //conditions for allowing the system to shoot: "fire" command received, weapon motors sufficiently fast, bullets loaded, last firing procedure completed
    trigger = 1; //debouncing the fire event: It cannot be triggered directly again in the next loop iteration
    weapontimer0 = millis();
  }

  weapontimer1 = millis(); 
  if (weapontimer0+3000>weapontimer1){ //safety limit: if the firing process is not completed after 3 s, the weapon is considered as jammed and auto-stops all motors
    if (trigger==1){   
      ledcWrite(BeltMotor,50); //duty cycle of the belt motor: be aware that its inertia may push it 1-2 more millimeters after stopping (more when moving faster)     
      if (belt==0){
        trigger = 2; //counting through the trigger levels: 1=firing starts; 2=belt has started moving (detected by photoelectric switch is not triggered by a belt pin)...
        weapondebounce = millis();
      }
    }else if (trigger==2&&weapontimer1>weapontimer0+100){ //debouncing the belt sensor by at least 100 ms between "no belt pin detected" and "belt pin detected"
      if (belt==1){
        trigger = 3; //...trigger=3 means the photoelectric switch is triggered (again) by a belt pin => it has moved to the next position
        ledcWrite(BeltMotor,0);
        pos = pos+1; //reduce the ammunition count by 1
        preferences.putInt("Ammunition", pos); //store ammunition count as a preference
        if(pos==2||pos==4||pos==6||pos==8||pos==10||pos==12||pos==14){ //tell the drum to move to the next position after every 2nd shot
          drummove = 1;  
        }
      }
    }
    
    if (trigger==3){
      if(drummove==1){
        ledcWrite(DrumMotor,45); //drum motor speed (duty cycle) governs how well the next position will be centered (also depending on the alignment of the drum sensor on the weapon)
        trigger = 4;
        weapondebounce = millis();
      }else{
        if (weapontimer1>weapontimer0+750){ //trigger debounce: new "fire" command is only accepted at least 0.75 s after the previous shot and if the drum is not moving
          trigger = 0;
        }
      }
    }
    
    if (trigger==4&&drum==1){ //preparation of drum stop when the old drum position was left
      drummove = 0;    
    }
    
    if (trigger==4&&drummove==0&&drum==0&&weapontimer1>weapondebounce+100){ //new drum position reached & 100 ms debounce criterion met
      ledcWrite(DrumMotor,0);
      trigger = 0;
    }
    
  }else{
    if (trigger>0){ //the weapon was triggered but the reloading was not completed within 3 s -> safety stop and jamming information
      ledcWrite(BeltMotor,0);
      ledcWrite(DrumMotor,0),
      pos = 17;
      preferences.putInt("Ammunition", pos); 
      trigger = 0;
      drummove = 0;
    }
  }  
}

//this function displays error information using the RGB LED strip
void RGBwarning(){
  if (outgoing.errorcode==3){ //error code 3: left motor train dysfunctional (gearbox transmission failed or motor stalled)
    strip.clear();
    for (int i=6; i<16; i++){ //let the left half of the RGB LEDs blink red; LED numbers depend on your wiring of the strip!
      strip.setPixelColor(i,127,0,0);
    }
    strip.show();
    delay(500);
    strip.clear();
    strip.show();
    delay(500);
  }
  if (outgoing.errorcode==4){ //error code 4: right motor train dysfunctional (gearbox transmission failed or motor stalled)
    strip.clear();
    for (int i=0; i<5; i++){ //let the right half of the RGB LEDs blink red; LED numbers depend on your wiring of the strip!
      strip.setPixelColor(i,127,0,0);
      strip.setPixelColor(i+17,127,0,0);
    }
    strip.show();
    delay(500);
    strip.clear();
    strip.show();
    delay(500);
  }
  if (outgoing.errorcode==2){ //error code 2: low battery error -> let the RGB LED strip send an SOS code with the center LED in front and back with low intensity
    for (int i=0; i<3; i++){ //"S" in the Morse code (3x short)
      strip.clear();
      strip.setPixelColor(5,30,0,0);
      strip.setPixelColor(16,30,0,0);
      strip.show();
      delay(400); //short = 0.4 s
      strip.clear();
      strip.show();
      delay(400); 
    }
    for (int i=0; i<3; i++){ //"O" in the Morse code (3x long)
      strip.clear();
      strip.setPixelColor(5,30,0,0);
      strip.setPixelColor(16,30,0,0);
      strip.show();
      delay(1200); //long = 1.2 s
      strip.clear();
      strip.show();
      delay(400); 
    }     
    for (int i=0; i<3; i++){ //"S" in the Morse code (3x short)
      strip.clear();
      strip.setPixelColor(5,30,0,0);
      strip.setPixelColor(16,30,0,0);
      strip.show();
      delay(400);
      strip.clear();
      strip.show();
      delay(400); 
    }
    delay(1000); //time between "words": 1.4 s        
  }
}

//the sentinel function is used to track inputs for possible errors
void sentinel(){
  if (pwml>55&&lspd<5){ //trigger limit for motor error counter: PWM duty cycle has to be >55 and the detected speed below 5 (in counts [arbitrary unit depending on code runtime])
    spderrL = spderrL +1;
  }else if (spderrL>0){
    spderrL=spderrL-1;
  }
  if (spderrL>1499){ //motor error counter at 1500 triggers the motor error (should be less than 1 s between stalling and error)
    outgoing.errorcode = 3;
  }
  if (pwmr>55&&rspd<5){
    spderrR = spderrR +1;
  }else if (spderrR>0){
    spderrR=spderrR-1;    
  }
  if (spderrR>1499){
    outgoing.errorcode = 4;
  }  
  if (batlim>99){ //low battery counter of 100 triggers the low battery error  
    outgoing.errorcode = 2;
  }
  if (outgoing.errorcode>0){ //error=1 means that the system goes into safety auto stop of all actuators within the main loop
    error = 1;
  }
}

//updatespeed0 is used for direct input to motor speed transmission
void updatespeed0(){
  if (incoming.spdL==0&&incoming.spdR==0){
    if (fantimer > 500){ //fantimer > 500 is reached after 15 s when not resetting the timer (which is done when the drive speed input is not equal to 0)
      ledcWrite(Fanchannel,0);
      fantimer = 500;
    }
  }else{
    fantimer = 0;
    ledcWrite(Fanchannel, 200);    
  }
  if (error==0){ //do only drive when there is no error
    if (incoming.spdL<0){ //conversion of drive inputs into PWM output: negative/positive refers to the direction; value refers to PWM duty cycle
      ledcWrite(EngPWM_L, (-incoming.spdL));
      digitalWrite(EngLDir, LOW);
    }else if(incoming.spdL>=0){
      ledcWrite(EngPWM_L, incoming.spdL);
      digitalWrite(EngLDir, HIGH);
    }
    pwml = abs(incoming.spdL); //transmitting the PWM duty cycle to this variable is needed to prevent abrupt changes when switching drive modes
    
    if (incoming.spdR<0){
      ledcWrite(EngPWM_R, (-incoming.spdR));
      digitalWrite(EngRDir, HIGH); 
    }else if(incoming.spdR>=0){
      ledcWrite(EngPWM_R, incoming.spdR);
      digitalWrite(EngRDir, LOW); 
    }
    pwmr = abs(incoming.spdR);
  }else{ //motors off when there is an error
    pwml=0;
    pwmr=0;
    ledcWrite(EngPWM_L, 0);
    ledcWrite(EngPWM_R, 0);    
  }
}

//updatespeed1 is responsible for the interpolated drive mode
void updatespeed1(){
  if (incoming.spdL==0&&incoming.spdR==0){ //fantimer works like in updatespeed0
    if (fantimer > 500){
      ledcWrite(Fanchannel,0);
      fantimer = 500;
    }
  }else{
    fantimer=0;
    ledcWrite(Fanchannel, 200);    
  }
  
  if (error==0){ //motor direction changes are executed below a PWM duty cycle of 10 (basically at standstill) - needs to be done here since transmitted speed and PWM value are not necessarily the same  
    if (incoming.spdL<0&&pwml<10){
      dirl = 0;
      digitalWrite(EngLDir, LOW);
    }else if(incoming.spdL>0&&pwml<10){
      dirl = 1;
      digitalWrite(EngLDir, HIGH);
    }
  
    if (dirl==0&&incoming.spdL<=0||dirl==1&&incoming.spdL>=0){ //safety precaution: speed changes relative to input speed are only done when they have the same direction
      if (pwml<abs(incoming.spdL)-30){ //simple interpolation logic: PWM duty cycle change increment size depends on the gap between input speed and current speed
        pwml = pwml+5;  
      }else if (pwml<abs(incoming.spdL)-10){
        pwml = pwml+3;
      }else if (pwml<abs(incoming.spdL)){
        pwml = pwml+1;
      }else if(pwml>abs(incoming.spdL)+30){
        pwml = pwml-5;
      }else if(pwml>abs(incoming.spdL)+10){
        pwml = pwml-3;  
      }else if(pwml>abs(incoming.spdL)&&pwml>0){
        pwml = pwml-1;
      }
    }else{
      pwml = pwml-7; //if speed input direction and system direction are not identical, the system reduces the speed quickly   
    }
    ledcWrite(EngPWM_L, pwml);
    
    if (incoming.spdR<0&&pwmr<10){ //the same for the right track
      dirr = 0;
      digitalWrite(EngRDir, HIGH); 
    }else if(incoming.spdR>0&&pwmr<10){
      dirr = 1;
      digitalWrite(EngRDir, LOW); 
    }
  
    if (dirr==0&&incoming.spdR<=0||dirr==1&&incoming.spdR>=0){
      if (pwmr<abs(incoming.spdR)-30){
        pwmr = pwmr+5;
      }else if (pwmr<abs(incoming.spdR)-10){
        pwmr = pwmr+3;
      }else if (pwmr<abs(incoming.spdR)){
        pwmr = pwmr+1;  
      }else if(pwmr>abs(incoming.spdR)+30){
        pwmr = pwmr-5;
      }else if(pwmr>abs(incoming.spdR)+10){
        pwmr = pwmr-3;  
      }else if(pwmr>abs(incoming.spdR)){
        pwmr = pwmr-1;
      }
    }else{
      pwmr = pwmr-7;   
    }
    ledcWrite(EngPWM_R, pwmr); 
  }else{ //safety stop when there is an error
    pwml = 0;
    pwmr = 0;
    ledcWrite(EngPWM_L, 0);
    ledcWrite(EngPWM_R, 0);  
  }
}

//Update of RGB LED strip
void updateRGB(){
   if (IntCount>=LEDSpeed){ //you can change the variable LEDspeed to adjust the speed of the running light (has to be changed in the code, not as a parameter in the remote control)
    IntCount = 0;
    if (wave==0){ //the combination of wave & counter determines the position and direction of the running light
      counter = counter+1; //the counter jumps to the next LED position after completing the fading (fading steps are encoded in the IntCount variable, which depends on LEDspeed)
    }else if(wave==1){
      counter = counter-1;
    }
  }

  if (wpnspd<150){ //the running light runs in its specified mode when the weapon is disabled. It automatically switches to the weapon indicator when the weapon is active
    IntMaxG     = incoming.G; //RGB codes are transmitted directly from the remote control
    IntMaxR     = incoming.R;
    IntMaxB     = incoming.B;
    currentIntG = map(IntCount,0,LEDSpeed-1,0,IntMaxG); //intensity fading is calculated based on the intensity counter value and the LEDspeed, which determines the number of fading steps
    currentIntR = map(IntCount,0,LEDSpeed-1,0,IntMaxR);
    currentIntB = map(IntCount,0,LEDSpeed-1,0,IntMaxB);
    IntCount    = IntCount+1;
    strip.clear();
    strip.setPixelColor(counter,currentIntR,currentIntG,currentIntB); //the current light has the increasing intensity values relative to IntMax
    strip.setPixelColor(counter+11,currentIntR,currentIntG,currentIntB); //11 lights in the front and 11 lights in the rear => code is running for both sides
  }else{ //weapon active means red low-intensity background light with more intense red running light
    IntMaxG     = 0;
    IntMaxR     = 127;
    IntMaxB     = 0;
    currentIntG = 0;
    currentIntR = map(IntCount,0,LEDSpeed-1,8,IntMaxR-8); //keeps the lower intensity limit for the running light at the level of the background light
    currentIntB = 0;
    IntCount    = IntCount+1;
    strip.clear();
    for (int i=0; i<RGBcount/2; i++){ //set all pixels to background color & intensity...
      strip.setPixelColor(i,8,0,0);
      strip.setPixelColor(i+11,8,0,0);
    }
    strip.setPixelColor(counter,currentIntR,currentIntG,currentIntB); //...and overwrite the current running light position with the respective intensity
    strip.setPixelColor(counter+11,currentIntR,currentIntG,currentIntB);    
  }
  
  if (wave==0){ //wave is responsible for the direction of the running light
    strip.setPixelColor(counter-1,IntMaxR,IntMaxG,IntMaxB); //the last light has the maximum intensity
    strip.setPixelColor(counter+10,IntMaxR,IntMaxG,IntMaxB); 
    strip.setPixelColor(counter-2,IntMaxR-currentIntR,IntMaxG-currentIntG,IntMaxB-currentIntB); //and the second last light has the decreasing intensity 
    strip.setPixelColor(counter+9,IntMaxR-currentIntR,IntMaxG-currentIntG,IntMaxB-currentIntB);    
  }else if (wave==1){
    strip.setPixelColor(counter+1,IntMaxR,IntMaxG,IntMaxB);  
    strip.setPixelColor(counter+12,IntMaxR,IntMaxG,IntMaxB);  
    strip.setPixelColor(counter+2,IntMaxR-currentIntR,IntMaxG-currentIntG,IntMaxB-currentIntB);  
    strip.setPixelColor(counter+13,IntMaxR-currentIntR,IntMaxG-currentIntG,IntMaxB-currentIntB);      
  }
  
  if (abs(incoming.spdL)+abs(incoming.spdR)>0){ //these if-conditions limit the running light boundaries to let the outer LEDs be used as head & rear lights when driving
    if (counter>=9&&IntCount>=LEDSpeed){
      wave     = 1;
      counter  = 7;
      IntCount = 0;
    }else if (counter<=1&&IntCount>=LEDSpeed){
      wave     = 0;
      counter  = 3;
      IntCount = 0;
    }
    if (incoming.spdL<0){ //the orientation of the LEDs depends on RGB strip wiring: needs to be adjusted manually accordingly
      strip.setPixelColor(10,127,0,0); //pixel on the front left: red light when driving backward
      strip.setPixelColor(11,abs(incoming.spdL),abs(incoming.spdL),abs(incoming.spdL)); //pixel on the rear left: white light when driving backward
    }else if(incoming.spdL>0){
      strip.setPixelColor(11,127,0,0); //other direction
      strip.setPixelColor(10,incoming.spdL,incoming.spdL,incoming.spdL); //the headlight intensity scales with speed
    }else{ 
      strip.setPixelColor(10,127,0,0); //red lights on the front and rear when only the right track is moving and the left is standing still
      strip.setPixelColor(11,127,0,0);  
    }
    if (incoming.spdR<0){
      strip.setPixelColor(0,127,0,0);
      strip.setPixelColor(21,abs(incoming.spdR),abs(incoming.spdR),abs(incoming.spdR));
    }else if (incoming.spdR>0){
      strip.setPixelColor(21,127,0,0);
      strip.setPixelColor(0,incoming.spdR,incoming.spdR,incoming.spdR);
    }else{
      strip.setPixelColor(0,127,0,0);
      strip.setPixelColor(21,127,0,0);           
    }
  }else{ //set counter & wave back when reaching the upper or lower limit
    if (counter>=10&&IntCount>=LEDSpeed){
      wave     = 1;
      counter  = 8;
      IntCount = 0;
    }else if (counter<=0&&IntCount>=LEDSpeed){
      wave     = 0;
      counter  = 2;
      IntCount = 0;
    }
  }
  strip.show();
}
